/*
   Copyright (c) 2019 by Plexim GmbH
   All rights reserved.

   A free license is granted to anyone to use this software for any legal
   non safety-critical purpose, including commercial applications, provided
   that:
   1) IT IS NOT USED TO DIRECTLY OR INDIRECTLY COMPETE WITH PLEXIM, and
   2) THIS COPYRIGHT NOTICE IS PRESERVED in its entirety.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
 */

#include "canbus.h"

CANBUS_Handle_t CANBUS_init(void *aMemory, const size_t aNumBytes){
	if(aNumBytes < sizeof(CANBUS_Obj_t))
	{
		return((CANBUS_Handle_t)NULL);
	}
	CANBUS_Handle_t handle = (CANBUS_Handle_t)aMemory;
	return handle;
}

void CANBUS_configure(CANBUS_Handle_t aHandle, uint16_t port, int32_t clk){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    obj->port = port;
    obj->clk = clk;
    obj->tseg1 = 0;
    obj->tseg2 = 0;
    obj->sjw = 1;
}

void CANBUS_setTiming(CANBUS_Handle_t aHandle, uint16_t tseg1, uint16_t tseg2, uint16_t sjw){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;
    obj->tseg1 = tseg1;
    obj->tseg2 = tseg2;
    obj->sjw = sjw;
}

bool CANBUS_setupPort(CANBUS_Handle_t aHandle, uint32_t baudrate){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    asm(" eallow");
    switch(obj->port)
    {
        case 0:
            // Launchpad
            obj->portHandle = CANA_BASE;

            CpuSysRegs.PCLKCR10.bit.CAN_A = 1;
            ClkCfgRegs.CLKSRCCTL2.bit.CANABCLKSEL = 0; // 0 = SYSCLK

            GpioCtrlRegs.GPBPUD.bit.GPIO33 = 0;     // enable pull-up for CANRXA
            GpioCtrlRegs.GPBPUD.bit.GPIO32 = 0;     // enable pull-up for CANTXA
            GpioCtrlRegs.GPBQSEL1.bit.GPIO33 = 3;   // asynch qual for CANRXA
            GpioCtrlRegs.GPBODR.bit.GPIO32 = 0;      // not open-drain (but rather push-pull)

            GpioCtrlRegs.GPBMUX1.bit.GPIO33 = 2;    // configure for CANRXA operation (mux 10)
            GpioCtrlRegs.GPBMUX1.bit.GPIO32 = 2;    // configure for CANTXA operation (mux 10)
            GpioCtrlRegs.GPBGMUX1.bit.GPIO33 = 2;
            GpioCtrlRegs.GPBGMUX1.bit.GPIO32 = 2;
            break;

        default:
             PLX_ASSERT(0);

    }
    asm(" edis");

    CAN_initModule(obj->portHandle);

    if((obj->tseg1 != 0) && (obj->tseg2 != 0))
    {
        uint32_t div = baudrate * (uint32_t)(obj->tseg1+obj->tseg2+1);
        uint16_t brr = (uint16_t)(obj->clk/div)-1;
        PLX_ASSERT(brr < 1024);
        PLX_ASSERT(((uint32_t)brr+1) * div == obj->clk); // too strict???
//      DEBUG_WRITE_STRING("BRR: 0x");
//      DEBUG_WRITE_INT(brr);
//      DEBUG_WRITE_STRING("\n\r");
        CAN_setBitTiming(obj->portHandle, (brr & 0x3F), (brr >> 6), obj->tseg1-1, obj->tseg2-1, obj->sjw);
    }
    else
    {
        uint16_t bitTime = 10; // must auto-determine proper value!
        CAN_setBitRate(obj->portHandle, obj->clk, baudrate, bitTime); // automatically sets timing for minimal latency
    }

//  uint32_t settings = CANBUS_getBitTiming(obj->portHandle);
//  DEBUG_WRITE_STRING("BTR: 0x");
//  DEBUG_WRITE_LONG(settings);
//  DEBUG_WRITE_STRING("\n\r");

    CAN_disableAutoBusOn(obj->portHandle); // note: if enabling, must also CANBUS_setAutoBusTime
    CAN_enableRetry(obj->portHandle);
    CAN_startModule(obj->portHandle); // clears INIT and CCE bits
    return true;
}

bool CANBUS_setupMailbox(CANBUS_Handle_t aHandle, uint16_t aMailBox, bool aConfigureForTransmit,
                         uint32_t aId, bool isExtended, unsigned char aLen)
{
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    CAN_MsgFrameType frame;
    uint32_t filter;

    if(isExtended)
    {
        frame = CAN_MSG_FRAME_EXT;
        filter = CAN_MSG_OBJ_USE_ID_FILTER | CAN_MSG_OBJ_USE_EXT_FILTER;
    }
    else
    {
        frame = CAN_MSG_FRAME_STD;
        filter = CAN_MSG_OBJ_USE_ID_FILTER;
    }

    if(aConfigureForTransmit)
    {
        CAN_setupMessageObject(obj->portHandle, aMailBox, aId, frame, CAN_MSG_OBJ_TYPE_TX,
                0, CAN_MSG_OBJ_NO_FLAGS, aLen);
    }
    else
    {
        CAN_setupMessageObject(obj->portHandle, aMailBox, aId, frame, CAN_MSG_OBJ_TYPE_RX,
                 0x1FFFFFFF, filter, aLen); // strict filtering; obj->rxLen does not seem to affect filtering!
    }
    return true;
}

void CANBUS_disableMailbox(CANBUS_Handle_t aHandle, uint16_t aMailBox){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    CAN_clearMessage(obj->portHandle, aMailBox);
}

bool CANBUS_isBusOn(CANBUS_Handle_t aHandle){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    if((CAN_getStatus(obj->portHandle) & CAN_STATUS_BUS_OFF) != 0) // check bit 7 of status register
    {
        return false;
    }
    else
    {
        return true;
    }
}

void CANBUS_setBusOn(CANBUS_Handle_t aHandle, bool busOn){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    if(busOn)
    {
        CAN_enableController(obj->portHandle); // clear INIT bit
    }
    else
    {
        CAN_disableController(obj->portHandle); // set INIT bit
    }
}

bool CANBUS_readyForTx(CANBUS_Handle_t aHandle, uint16_t aMailBox){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    if((CAN_getStatus(obj->portHandle) & CAN_STATUS_BUS_OFF) != 0)
    {
        CAN_enableController(obj->portHandle); // attempt to go online again
        return false;
    }

    uint32_t mask = 1 << (aMailBox-1);
    if((CAN_getTxRequests(obj->portHandle) &  mask) != 0) // CAN_TXRQ_21
    {
        return false;
    }
    else
    {
        return true;
    }
}

bool CANBUS_getMessage(CANBUS_Handle_t aHandle, uint16_t aMailBox, unsigned char data[], unsigned char *len){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    uint16_t configuredLen = HWREGH(obj->portHandle + CAN_O_IF1MCTL) & CAN_IF1MCTL_DLC_M;

    uint32_t mask = 1 << (aMailBox-1);
    if((CAN_getNewDataFlags(obj->portHandle) &  mask) == 0) // CAN_NDAT_21
    {
        return false; // no new data
    }
    else
    {
        CAN_readMessage(obj->portHandle, aMailBox, (uint16_t *)data);
        *len = (unsigned char)configuredLen;
        return true;
    }
}

bool CANBUS_putMessage(CANBUS_Handle_t aHandle, uint16_t aMailBox, unsigned char data[], unsigned char len){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    uint16_t configuredLen = HWREGH(obj->portHandle + CAN_O_IF1MCTL) & CAN_IF1MCTL_DLC_M;

    if(len != (unsigned char)configuredLen)
    {
        // catch here, b/c CAN_sendMessage has an ASSERT on this condition
        return false;
    }

    uint32_t mask = 1 << (aMailBox-1);
    if((CAN_getTxRequests(obj->portHandle) &  mask) != 0) // CAN_TXRQ_21
    {
        return false; // still busy!
    }
    else
    {
        CAN_sendMessage(obj->portHandle, aMailBox, len, (uint16_t*)data);
        return true;
    }
}

