/*
   Copyright (c) 2019 by Plexim GmbH
   All rights reserved.

   A free license is granted to anyone to use this software for any legal
   non safety-critical purpose, including commercial applications, provided
   that:
   1) IT IS NOT USED TO DIRECTLY OR INDIRECTLY COMPETE WITH PLEXIM, and
   2) THIS COPYRIGHT NOTICE IS PRESERVED in its entirety.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
 */

#include "canbus.h"

CANBUS_Handle_t CANBUS_init(void *aMemory, const size_t aNumBytes){
	if(aNumBytes < sizeof(CANBUS_Obj_t))
	{
		return((CANBUS_Handle_t)NULL);
	}
	CANBUS_Handle_t handle = (CANBUS_Handle_t)aMemory;
	return handle;
}

void CANBUS_configure(CANBUS_Handle_t aHandle, uint16_t port, int32_t clk){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    obj->port = port;
    obj->clk = clk;
    obj->tseg1 = 0;
    obj->tseg2 = 0;
    obj->sjw = 1;
}

void CANBUS_setTiming(CANBUS_Handle_t aHandle, uint16_t tseg1, uint16_t tseg2, uint16_t sjw){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;
    obj->tseg1 = tseg1;
    obj->tseg2 = tseg2;
    obj->sjw = sjw;
}

bool CANBUS_setupPort(CANBUS_Handle_t aHandle, uint32_t baudrate){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    asm(" eallow");
    switch(obj->port)
    {
    case 0:
        // ControlCard
        obj->portHandle = CANA_BASE;

        CpuSysRegs.PCLKCR10.bit.CAN_A = 1;
        ClkCfgRegs.CLKSRCCTL2.bit.CANABCLKSEL = 0;

        GpioCtrlRegs.GPAPUD.bit.GPIO30 = 0;     // Enable pull-up for GPIO30 (CANRXA)
        GpioCtrlRegs.GPAPUD.bit.GPIO31 = 0;     // Enable pull-up for GPIO31 (CANTXA)
        GpioCtrlRegs.GPAQSEL2.bit.GPIO30 = 3;   // Asynch qual for GPIO30 (CANRXA)
        GpioCtrlRegs.GPAODR.bit.GPIO31 = 0;      // GPIO31 not open-drain (but rather push-pull)

        GpioCtrlRegs.GPAMUX2.bit.GPIO30 = 1;    // Configure GPIO30 for CANRXA operation
        GpioCtrlRegs.GPAGMUX2.bit.GPIO30 = 0;
        GpioCtrlRegs.GPAMUX2.bit.GPIO31 = 1;    // Configure GPIO31 for CANTXA operation
        GpioCtrlRegs.GPAGMUX2.bit.GPIO31 = 0;
        break;

    case 2:
        // Launchpad 28377S
        obj->portHandle = CANA_BASE;

        CpuSysRegs.PCLKCR10.bit.CAN_A = 1;
        ClkCfgRegs.CLKSRCCTL2.bit.CANABCLKSEL = 0;

        GpioCtrlRegs.GPCPUD.bit.GPIO70 = 0;     // Enable pull-up for GPIO70 (CANRXB)
        GpioCtrlRegs.GPCPUD.bit.GPIO71 = 0;     // Enable pull-up for GPIO71 (CANTXB)
        GpioCtrlRegs.GPCQSEL1.bit.GPIO70 = 3;   // Asynch qual for GPIO70 (CANRXB)
        GpioCtrlRegs.GPCODR.bit.GPIO71 = 0;      // GPIO71 not open-drain (but rather push-pull)

        GpioCtrlRegs.GPCMUX1.bit.GPIO70 = 1;    // Configure GPIO70 for CANRXB operation
        GpioCtrlRegs.GPCGMUX1.bit.GPIO70 = 1;
        GpioCtrlRegs.GPCMUX1.bit.GPIO71 = 1;    // Configure GPIO71 for CANTXB operation
        GpioCtrlRegs.GPCGMUX1.bit.GPIO71 = 1;
        break;

    case 10:
        // Launchpad 28379D
        obj->portHandle = CANB_BASE;

        CpuSysRegs.PCLKCR10.bit.CAN_B = 1;
        ClkCfgRegs.CLKSRCCTL2.bit.CANABCLKSEL = 0;

        GpioCtrlRegs.GPAPUD.bit.GPIO17 = 0;     // Enable pull-up for GPIO17 (CANRXB)
        GpioCtrlRegs.GPAPUD.bit.GPIO12 = 0;     // Enable pull-up for GPIO12 (CANTXB)
        GpioCtrlRegs.GPAQSEL2.bit.GPIO17 = 3;   // Asynch qual for GPIO17 (CANRXB)
        GpioCtrlRegs.GPAODR.bit.GPIO12 = 0;      // GPIO12 not open-drain (but rather push-pull)

        GpioCtrlRegs.GPAMUX2.bit.GPIO17 = 2;    // Configure GPIO17 for CANRXB operation
        GpioCtrlRegs.GPAGMUX2.bit.GPIO17 = 0;
        GpioCtrlRegs.GPAMUX1.bit.GPIO12 = 2;    // Configure GPIO12 for CANTXB operation
        GpioCtrlRegs.GPAGMUX1.bit.GPIO12 = 0;
        break;

        default:
             PLX_ASSERT(0);

    }
    asm(" edis");

    CANInit(obj->portHandle);
    CANBitRateSet(obj->portHandle, obj->clk, baudrate); // automatically sets timing for minimal latency
    CANEnable(obj->portHandle);
    return true;
}

bool CANBUS_setupMailbox(CANBUS_Handle_t aHandle, uint16_t aMailBox, bool aConfigureForTransmit,
                         uint32_t aId, bool isExtended, unsigned char aLen)
{
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    if(aConfigureForTransmit)
    {
        obj->msgObj[aMailBox].ui32MsgID = aId;
        obj->msgObj[aMailBox].ui32MsgIDMask = 0;
        obj->msgObj[aMailBox].ui32Flags = 0;
        if(isExtended){
            obj->msgObj[aMailBox].ui32Flags |= MSG_OBJ_EXTENDED_ID;
        }
        obj->msgObj[aMailBox].ui32MsgLen = 8;
    }
    else
    {
        obj->msgObj[aMailBox].ui32MsgID = aId;
        obj->msgObj[aMailBox].ui32MsgIDMask = 0xFFFFFFFF;
        obj->msgObj[aMailBox].ui32Flags = MSG_OBJ_USE_ID_FILTER;
        if(isExtended){
            obj->msgObj[aMailBox].ui32Flags |= MSG_OBJ_EXTENDED_ID;
        }
        obj->msgObj[aMailBox].ui32MsgLen = 8;
        CANMessageSet(obj->portHandle, aMailBox, &obj->msgObj[aMailBox], MSG_OBJ_TYPE_RX);
    }
    return true;
}

void CANBUS_disableMailbox(CANBUS_Handle_t aHandle, uint16_t aMailBox){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    CANMessageClear(obj->portHandle, aMailBox);
}

bool CANBUS_isBusOn(CANBUS_Handle_t aHandle){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    if((CANStatusGet(obj->portHandle, CAN_STS_CONTROL) & CAN_STATUS_BUS_OFF) != 0){
        return false;
    } else {
        return true;
    }
}

void CANBUS_setBusOn(CANBUS_Handle_t aHandle, bool busOn){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    if(busOn){
        CANEnable(obj->portHandle);
    } else {
        CANDisable(obj->portHandle);
    }
}

bool CANBUS_readyForTx(CANBUS_Handle_t aHandle, uint16_t aMailBox){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    if((CANStatusGet(obj->portHandle, CAN_STS_CONTROL) & CAN_STATUS_BUS_OFF) != 0){
        CANEnable(obj->portHandle); // attempt to go online again
        return false;
    }

    uint32_t mask = 1 << (aMailBox-1);
    if((CANStatusGet(obj->portHandle, CAN_STS_TXREQUEST) & mask) != 0){
        return false;
    } else {
        return true;
    }
}

bool CANBUS_getMessage(CANBUS_Handle_t aHandle, uint16_t aMailBox, unsigned char data[], unsigned char *len){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    uint32_t mask = 1 << (aMailBox-1);
    if((CANStatusGet(obj->portHandle, CAN_STS_NEWDAT) & mask) == 0){
        return false;
    } else {
        obj->msgObj[aMailBox].pucMsgData = data;
        CANMessageGet(obj->portHandle, aMailBox, &obj->msgObj[aMailBox], true);
        *len = (unsigned char)obj->msgObj[aMailBox].ui32MsgLen;
        return true;
    }
}

bool CANBUS_putMessage(CANBUS_Handle_t aHandle, uint16_t aMailBox, unsigned char data[], unsigned char len){
    CANBUS_Obj_t *obj = (CANBUS_Obj_t *)aHandle;

    uint32_t mask = 1 << (aMailBox-1);
    if((CANStatusGet(obj->portHandle, CAN_STS_TXREQUEST) & mask) != 0){
        return false;
    } else {
        obj->msgObj[aMailBox].pucMsgData = data;
        obj->msgObj[aMailBox].ui32MsgLen = len;
        CANMessageSet(obj->portHandle, aMailBox, &obj->msgObj[aMailBox], MSG_OBJ_TYPE_TX);
        return true;
    }
}

