/*
   Copyright (c) 2014-2019 by Plexim GmbH
   All rights reserved.

   A free license is granted to anyone to use this software for any legal
   non safety-critical purpose, including commercial applications, provided
   that:
   1) IT IS NOT USED TO DIRECTLY OR INDIRECTLY COMPETE WITH PLEXIM, and
   2) THIS COPYRIGHT NOTICE IS PRESERVED in its entirety.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
 */

#include "power.h"
#include "pil.h"

#pragma diag_suppress 179 // enter_PS_FSM_STATE_CRITICAL_FAULT declared but not referenced

#define PS_ENABLE_DELAY_MS 100

PWR_Obj_t PWR_SObj;
PWR_Handle_t PWR_SHandle;

static bool PWR_isSafe();
static void PWR_disableSwitching();
static bool PWR_hasTripped();
static void PWR_reset();

void PWR_sinit()
{
    PWR_SHandle = (PWR_Handle_t)&PWR_SObj;
}

void PWR_configure(GDRV_Handle_t aGateDrvHandle, uint16_t aFsmExecRateHz)
{
    PWR_Obj_t *obj = (PWR_Obj_t *)PWR_SHandle;
    obj->gdrvHandle = aGateDrvHandle;

    obj->fsmExecRateHz = aFsmExecRateHz;

    PWR_unregisterAllPwmChannels();

    obj->pilMode = false;
    obj->gdrvInitReq = true;
    obj->gdrvStatus = 0;

    GDRV_powerup(obj->gdrvHandle);
    PWR_reset();
}

void PWR_unregisterAllPwmChannels()
{
    PWR_Obj_t *obj = (PWR_Obj_t *)PWR_SHandle;

    obj->numRegisteredPwmChannels = 0;
}

void PWR_registerPwmChannel(PWM_Handle_t aChannel)
{
    PWR_Obj_t *obj = (PWR_Obj_t *)PWR_SHandle;

    PLX_ASSERT(obj->numRegisteredPwmChannels < PWR_MAX_PWM_CHANNELS);
    obj->pwmChannels[obj->numRegisteredPwmChannels] = aChannel;
    obj->numRegisteredPwmChannels++;
}

void PWR_reset()
{
    PWR_Obj_t *obj = (PWR_Obj_t *)PWR_SHandle;

    obj->gatesActive = false;
    obj->enableSwitchingReq = false;
    obj->state = PWR_STATE_POWERUP;
    obj->enableReq = false;
    obj->errorCode = PWR_ERR_NONE;
}

bool PWR_isReadyForEnable()
{
    PWR_Obj_t *obj = (PWR_Obj_t *)PWR_SHandle;

    DINT;
    bool isReadyForEnable = (obj->state == PWR_STATE_DISABLED);
    EINT;
    return isReadyForEnable;
}

void PWR_enable(bool aEnable)
{
    PWR_Obj_t *obj = (PWR_Obj_t *)PWR_SHandle;

    DINT;
    obj->enableReq = aEnable;
    EINT;
}

bool PWR_isEnabled()
{
    PWR_Obj_t *obj = (PWR_Obj_t *)PWR_SHandle;
    DINT;
    bool isEnabled = (obj->state == PWR_STATE_ENABLED);
    EINT;
    return isEnabled;
}

void PWR_setPilMode(bool pilMode)
{
    PWR_Obj_t *obj = (PWR_Obj_t *)PWR_SHandle;

    DINT;
    obj->pilMode = pilMode;
    EINT;
}

void PWR_fsm()
{
    PWR_Obj_t *obj = (PWR_Obj_t *)PWR_SHandle;

    if(obj->gdrvStatus <= 0)
    {
        // gate-driver not ready
        return;
    }

    DINT; // disable interrupts to allow changing enableReq from highest priority task
    uint16_t enableReq = obj->enableReq;
    EINT;
    int16_t newState = obj->state;
    switch(obj->state)
    {
        // powerup
        case PWR_STATE_POWERUP:
            if(!enableReq)
            {
                goto enter_PS_FSM_STATE_DISABLED;
            }
            break;

        // disabled
        enter_PS_FSM_STATE_DISABLED:
            GDRV_disable(obj->gdrvHandle);
            newState = PWR_STATE_DISABLED;
            PWR_disableSwitching();

        break;
        case PWR_STATE_DISABLED:
            if(enableReq)
            {
                goto enter_PS_FSM_STATE_ENABLING;
            }
            break;

        // enabling
        enter_PS_FSM_STATE_ENABLING:
            newState = PWR_STATE_ENABLING;
            if(obj->pilMode != true)
            {
                GDRV_enable(obj->gdrvHandle);
                obj->timer = obj->fsmExecRateHz/1000 * PS_ENABLE_DELAY_MS; // for charge-pump to reach steady-state
                if(obj->timer < 1)
                {
                    obj->timer = 1; // enforce minimal delay
                }
            }
            else
            {
                obj->timer = 0;
            }
            break;
        case PWR_STATE_ENABLING:
            if(obj->timer == 0)
            {
                goto enter_PS_FSM_STATE_ENABLED;
            }
            else
            {
                obj->timer--;
            }
            break;

        // enabled
        enter_PS_FSM_STATE_ENABLED:
            // enable powerstage
            newState = PWR_STATE_ENABLED;
            DINT;
            obj->enableSwitchingReq = true;
            EINT;
        break;
        case PWR_STATE_ENABLED:
            // check for fault
            if(GDRV_hasFaulted(obj->gdrvHandle))
            {
                goto enter_PS_FSM_STATE_FAULT;
            }
            if(!enableReq)
            {
                goto enter_PS_FSM_STATE_DISABLED;
            }
            else
            {
                if(PWR_hasTripped())
                {
                    goto enter_PS_FSM_STATE_FAULT;
                }
            }
            break;

        // fault
        enter_PS_FSM_STATE_FAULT:
            PWR_disableSwitching();
            GDRV_disable(obj->gdrvHandle);
            newState = PWR_STATE_FAULT;
        break;
        case PWR_STATE_FAULT:
            if(!enableReq)
            {
                goto enter_PS_FSM_STATE_FAULT_ACKN;
            }
            break;

        // critical fault
        enter_PS_FSM_STATE_CRITICAL_FAULT:
            PWR_disableSwitching();
            GDRV_disable(obj->gdrvHandle);
            newState = PWR_STATE_CRITICAL_FAULT;
        break;
        case PWR_STATE_CRITICAL_FAULT:
            // no way out
            break;

        // fault acknowledge
        enter_PS_FSM_STATE_FAULT_ACKN:
            newState = PWR_STATE_FAULT_ACKN;
            break;
        case PWR_STATE_FAULT_ACKN:
            // reset gate driver here...
            goto enter_PS_FSM_STATE_DISABLED;
    }
    DINT; // disable interrupts to allow reading state from higher priority tasks
    obj->state = newState;
    EINT;
}

void PWR_background()
{
    PWR_Obj_t *obj = (PWR_Obj_t *)PWR_SHandle;

    if(obj->gdrvInitReq)
    {
        if(GDRV_setup(obj->gdrvHandle))
        {
            obj->gdrvStatus = 1;
        }
        else
        {
            obj->gdrvStatus = -1;
        }
        obj->gdrvInitReq = false;
    }
}

static bool PWR_isSafe()
{
    PWR_Obj_t *obj = (PWR_Obj_t *)PWR_SHandle;

    return ((obj->state != PWR_STATE_ENABLED) && (obj->state != PWR_STATE_ENABLING));
}

static void PWR_disableSwitching()
{
    PWR_Obj_t *obj = (PWR_Obj_t *)PWR_SHandle;

    DINT;
    obj->enableSwitchingReq = false;
    EINT;

    // disable actuators
    int i;
    for(i=0; i< obj->numRegisteredPwmChannels; i++)
    {
        PWM_disableOut(obj->pwmChannels[i]);
    }
    obj->gatesActive = false;
}

static bool PWR_hasTripped()
{
    PWR_Obj_t *obj = (PWR_Obj_t *)PWR_SHandle;
    if(obj->pilMode){
        return false;
    }
    int i;
    for(i=0; i< obj->numRegisteredPwmChannels; i++)
    {
        if(obj->pwmChannels[i]->pwm->TZFLG.bit.OST == 1)
        {
            return true;
        }
    }
    return false;
}
