/*
   Copyright (c) 2014-2019 by Plexim GmbH
   All rights reserved.

   A free license is granted to anyone to use this software for any legal
   non safety-critical purpose, including commercial applications, provided
   that:
   1) IT IS NOT USED TO DIRECTLY OR INDIRECTLY COMPETE WITH PLEXIM, and
   2) THIS COPYRIGHT NOTICE IS PRESERVED in its entirety.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
 */

#include "dispatcher.h"
#include "f28004x_epwm_defines.h"

/*
 * Assumptions/simplifications:
 * 1) ADC configured to trigger ADCINT1
 * 2) ePWM configured to issue SOCA
 * 3) ePWM int on PIE group 3 (ok for epwm1-8)
 * 4) ePWM and CpuTimer0 clock = SYSCLK/2
 * 5) Use of CpuTimer0 is hardcoded
 */

// singleton (for now)
DISPR_Obj_t DisprObj;
DISPR_Handle_t DisprHandle;

// for diagnostics purposes
#ifndef PIL_PREP_TOOL
PIL_SYMBOL_DEF(DisprObj_ctrlPeriod, 0, 1.0, "");
PIL_SYMBOL_DEF(DisprObj_task2DispatchPeriod, 0, 1.0, "");
PIL_SYMBOL_DEF(DisprObj_thisCntdownTimestamp0, 0, 1, "");
PIL_SYMBOL_DEF(DisprObj_thisCntdownTimestamp1, 0, 1, "");
PIL_SYMBOL_DEF(DisprObj_thisCntdownTimestamp2, 0, 1, "");
PIL_SYMBOL_DEF(DisprObj_cntdownTimestamp1, 0, 1, "");
PIL_SYMBOL_DEF(DisprObj_cntdownTimestamp2, 0, 1, "");
PIL_SYMBOL_DEF(DisprObj_task2Active, 0, 1, "");
PIL_SYMBOL_DEF(DisprObj_task2Timestamp, 0, 1, "");
PIL_SYMBOL_DEF(DisprObj_thisTask2Timestamp, 0, 1, "");
PIL_SYMBOL_DEF(DisprObj_task1Margin, 0, 1, "[1]");
PIL_SYMBOL_DEF(DisprObj_task2Margin, 0, 1, "[1]");
PIL_SYMBOL_DEF(DisprObj_task1WorstCaseMargin, 0, 1, "[1]");
PIL_SYMBOL_DEF(DisprObj_task2WorstCaseMargin, 0, 1, "[1]");
#endif

static interrupt void DISPR_dispatch();

void DISPR_sinit()
{
    DisprHandle = (DISPR_Handle_t)&DisprObj;

    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;
    obj->powerupDelayIntTask1Ticks = 0;
    obj->timeBase =  DISPR_TimebaseIsUndefined;
}

void DISPR_configure(uint32_t aSysClkHz, PIL_Handle_t aPilHandle)
{
    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;

    obj->sysClkHz = aSysClkHz;
    obj->pilHandle = aPilHandle;
}

static void DISPR_configurePwm(PWM_Unit_t aPwmChannel)
{
    PLX_ASSERT((int)aPwmChannel >= (int)PWM_EPWM1);
    PLX_ASSERT((int)aPwmChannel <= (int)PWM_EPWM8);

    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;
    PWM_getRegisterBase(aPwmChannel, &obj->ePwm);
    obj->ePwmIntVector = (PINT *)((uint32_t)(&PieVectTable.EPWM1_INT) + ((uint32_t)aPwmChannel-1)*sizeof(PINT *));
    obj->ePwmIerBitMask = (1 << (aPwmChannel-1));
}

static void DISPR_configureAdc(AIN_Unit_t aAdcUnit)
{
    PLX_ASSERT((int)aAdcUnit >= (int)AIN_ADC_A);
    PLX_ASSERT((int)aAdcUnit <= (int)AIN_ADC_C);

    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;
    AIN_getRegisterBase(aAdcUnit, &obj->adc);

    obj->adcIntVector = (PINT *)((uint32_t)(&PieVectTable.ADCA1_INT) + ((uint32_t)aAdcUnit)*sizeof(PINT *));
    obj->adcPie1BitMask = (1 << aAdcUnit);
}

void DISPR_setPowerupDelay(uint16_t aTask1Ticks)
{
    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;
    obj->powerupDelayIntTask1Ticks = aTask1Ticks;
}

bool DISPR_powerupDelayIsActive()
{
    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;

    return (obj->powerupCountdown > 0);
}

void DISPR_setTriggerByAdcViaTimer(AIN_Unit_t aAdcUnit, uint16_t aEocChannel)
{
    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;

    DISPR_stop();

    DISPR_configureAdc(aAdcUnit);
    obj->timeBase = DISPR_TimebaseIsAdc;
    obj->baseClock = DISPR_BaseClockIsTimer;

    // enable trigger to SOC
    CpuTimer0Regs.TCR.bit.TIE = 1;

    EALLOW;
    obj->adc->ADCINTSEL1N2.bit.INT1CONT = 0; // disable ADCINT1 Continuous mode
    obj->adc->ADCINTSEL1N2.bit.INT1SEL = aEocChannel; // setup EOC to trigger ADCINT1
    obj->adc->ADCINTSEL1N2.bit.INT1E = 1; // enable ADCINT1
    obj->adc->ADCCTL1.bit.INTPULSEPOS = 1; // ADCINT1 trips after AdcResults latch
    *obj->adcIntVector = &DISPR_dispatch;
    EDIS;
}

void DISPR_setTriggerByAdcViaPwm(AIN_Unit_t aAdcUnit, uint16_t aEocChannel, PWM_Unit_t aPwmChannel)
{
    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;

    DISPR_stop();

    DISPR_configurePwm(aPwmChannel);

    DISPR_configureAdc(aAdcUnit);
    obj->timeBase = DISPR_TimebaseIsAdc;
    obj->baseClock = DISPR_BaseClockIsPwm;

    // work here: should also work without PWM, should make sure only one SOCAEN is set
    obj->ePwm->ETSEL.bit.SOCAEN = 1; // enable SOC on A group

    EALLOW;
    obj->adc->ADCINTSEL1N2.bit.INT1CONT = 0; // disable ADCINT1 Continuous mode
    obj->adc->ADCINTSEL1N2.bit.INT1SEL = aEocChannel; // setup EOC to trigger ADCINT1
    obj->adc->ADCINTSEL1N2.bit.INT1E = 1; // enable ADCINT1
    obj->adc->ADCCTL1.bit.INTPULSEPOS = 1; // ADCINT1 trips after AdcResults latch
    *obj->adcIntVector = &DISPR_dispatch;
    EDIS;
}

void DISPR_setTriggerByTimer()
{
    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;

    DISPR_stop();

    obj->timeBase = DISPR_TimebaseIsTimer;
    obj->baseClock = DISPR_BaseClockIsTimer;

    PieCtrlRegs.PIEIER1.bit.INTx7 = 1;
    CpuTimer0Regs.TCR.bit.TIE = 1;

    EALLOW;
    PieVectTable.TIMER0_INT = &DISPR_dispatch;
    EDIS;
}

void DISPR_setTriggerByPwm(PWM_Unit_t aPwmChannel, uint16_t aEventSel)
{
    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;

    DISPR_stop();

    DISPR_configurePwm(aPwmChannel);

    obj->timeBase = DISPR_TimebaseIsPwm;
    obj->baseClock = DISPR_BaseClockIsPwm;

    obj->ePwm->ETSEL.bit.INTSEL = aEventSel;
    obj->ePwm->ETSEL.bit.INTEN = 1;  // enable INT

    PieCtrlRegs.PIEIER3.all |= obj->ePwmIerBitMask;

    EALLOW;
    *obj->ePwmIntVector = &DISPR_dispatch;
    EDIS;
}

void DISPR_reset()
{
    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;
    obj->task2DispatchTimer = 0;
}

void DISPR_start(DISPR_CallbackPtr_t aTsk1, uint32_t aTsk1PeriodInSysClkTicks,
        DISPR_CallbackPtr_t aTsk2, uint32_t aTsk2PeriodInSysClkTicks)
{
    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;

    obj->tsk1 = aTsk1;
    obj->tsk2 = aTsk2;

    uint32_t epwmClkPeriodInSysClkTicks = 2; // 1/2 system clock

    CpuTimer0Regs.TCR.bit.TSS = 1; // stop timer
    CpuTimer0Regs.TPRH.all = 0;
    CpuTimer0Regs.TPR.all = 1; // set cpu timer to same clock as ePWM (SYSCLK/2)

    if(obj->baseClock == DISPR_BaseClockIsTimer)
    {
        obj->ctrlPeriod =  aTsk1PeriodInSysClkTicks/epwmClkPeriodInSysClkTicks; // timer at ePWM clock
        PLX_ASSERT(aTsk1PeriodInSysClkTicks == epwmClkPeriodInSysClkTicks*obj->ctrlPeriod);

        CpuTimer0Regs.PRD.all = obj->ctrlPeriod-1;
        CpuTimer0Regs.TCR.bit.TRB = 1; // reload timer
    }
    else if(obj->baseClock == DISPR_BaseClockIsPwm)
    {
        PLX_ASSERT(obj->ePwm->TBCTL.bit.CTRMODE != TB_COUNT_DOWN);

        bool firstTriggerOnPeriod = false;
        bool doubleEdgeTrigger = false;
        uint32_t intClkPeriodInSysClkTicks;
        if(obj->ePwm->TBCTL.bit.CTRMODE == TB_COUNT_UPDOWN){
            // triangle
            intClkPeriodInSysClkTicks = epwmClkPeriodInSysClkTicks*2L*(uint32_t)obj->ePwm->TBPRD;
            if(obj->timeBase == DISPR_TimebaseIsAdc){
                if(obj->ePwm->ETSEL.bit.SOCASEL == ET_CTR_PRDZERO){
                    doubleEdgeTrigger = true;
                } else {
                    firstTriggerOnPeriod = (obj->ePwm->ETSEL.bit.SOCASEL == ET_CTR_PRD);
                }
             } else {
                if(obj->ePwm->ETSEL.bit.INTSEL == ET_CTR_PRDZERO){
                    doubleEdgeTrigger = true;
                } else {
                    firstTriggerOnPeriod = (obj->ePwm->ETSEL.bit.INTSEL == ET_CTR_PRD);
                }
            }
        } else {
            // sawtooth
            intClkPeriodInSysClkTicks = epwmClkPeriodInSysClkTicks*(uint32_t)(obj->ePwm->TBPRD+1);
        }

        if(doubleEdgeTrigger){
            PLX_ASSERT(obj->ePwm->TBCTL.bit.CTRMODE == TB_COUNT_UPDOWN);
            intClkPeriodInSysClkTicks = intClkPeriodInSysClkTicks/2;
        }

        uint16_t prd = (uint16_t)(aTsk1PeriodInSysClkTicks/intClkPeriodInSysClkTicks);
        if(((uint32_t)prd*intClkPeriodInSysClkTicks) != aTsk1PeriodInSysClkTicks)
        {
            PLX_ASSERT(0); // only exact multiples allowed
        }
        if(prd > 15) // type D, up to 15
        {
            PLX_ASSERT(0); // ratio not supported
        }

        if(obj->timeBase == DISPR_TimebaseIsAdc)
        {
            obj->ePwm->ETPS.bit.SOCPSSEL = 1;
            obj->ePwm->ETSOCPS.bit.SOCAPRD2 = prd;
        }
        else
        {
            obj->ePwm->ETPS.bit.INTPSSEL = 1;
            obj->ePwm->ETINTPS.bit.INTPRD2 = prd;
        }

        // configure timer for timing diagnostics
        if(obj->ePwm->TBCTL.bit.CTRMODE == TB_COUNT_UPDOWN){
            // triangle
            obj->ctrlPeriod =  (uint32_t)prd * (uint32_t)obj->ePwm->TBPRD;
            if(!doubleEdgeTrigger){
                obj->ctrlPeriod = obj->ctrlPeriod * 2;
            }
            if(firstTriggerOnPeriod){
                CpuTimer0Regs.PRD.all = obj->ctrlPeriod - 1 - (uint32_t)obj->ePwm->TBPRD; // initial offset
                CpuTimer0Regs.TCR.bit.TRB = 1; // reload timer (initial offset)
                CpuTimer0Regs.PRD.all = obj->ctrlPeriod-1;
            } else {
                CpuTimer0Regs.PRD.all = obj->ctrlPeriod-1;
                CpuTimer0Regs.TCR.bit.TRB = 1; // reload timer
            }
        } else {
            // sawtooth
            obj->ctrlPeriod =  (uint32_t)prd * (uint32_t)(obj->ePwm->TBPRD+1);
            CpuTimer0Regs.PRD.all = obj->ctrlPeriod-1;
            CpuTimer0Regs.TCR.bit.TRB = 1; // reload timer
        }
    }

    obj->task2DispatchPeriod = (uint16_t)(aTsk2PeriodInSysClkTicks/aTsk1PeriodInSysClkTicks);
    if(((uint32_t)obj->task2DispatchPeriod*aTsk1PeriodInSysClkTicks) != aTsk2PeriodInSysClkTicks)
    {
        PLX_ASSERT(0); // only exact multiples allowed
    }

    obj->task2Active = false;
    obj->interruptNesting = 0;
    obj->cntdownTimestamp1 = 0xFFFFFFFF;
    obj->cntdownTimestamp2 = 0xFFFFFFFF;
    obj->task2Timestamp = 0;
    DISPR_reset();

    obj->powerupCountdown = obj->powerupDelayIntTask1Ticks;

    if(obj->timeBase == DISPR_TimebaseIsAdc)
    {
        PieCtrlRegs.PIEIER1.all |= obj->adcPie1BitMask;
        PieCtrlRegs.PIEACK.all = PIEACK_GROUP1; // Acknowledge interrupt to PIE (all ADCs in group 1)
        IER |= M_INT1;
    }
    else if (obj->timeBase == DISPR_TimebaseIsPwm)
    {
        PieCtrlRegs.PIEIER3.all |= obj->ePwmIerBitMask;
        PieCtrlRegs.PIEACK.all = PIEACK_GROUP3; // Acknowledge interrupt to PIE
        IER |= M_INT3;
    }
    else // timer
    {
        PieCtrlRegs.PIEIER1.bit.INTx7 = 1;
        PieCtrlRegs.PIEACK.all = PIEACK_GROUP1; // Acknowledge interrupt to PIE
        IER |= M_INT1;
    }
}

void DISPR_stop()
{
    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;

    switch(obj->timeBase){
        default:
        case DISPR_TimebaseIsUndefined:
            break;

        case DISPR_TimebaseIsTimer:
            PieCtrlRegs.PIEIER1.bit.INTx7 = 0;
            break;

        case DISPR_TimebaseIsPwm:
            PieCtrlRegs.PIEIER3.all &= ~obj->ePwmIerBitMask;
            break;

        case DISPR_TimebaseIsAdc:
            PieCtrlRegs.PIEIER1.all &= ~obj->adcPie1BitMask;
            break;
    }
}

static interrupt void DISPR_dispatch(){
    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;

    obj->thisCntdownTimestamp0 = CpuTimer0Regs.TIM.all;

    if(obj->timeBase == DISPR_TimebaseIsAdc)
    {
        obj->adc->ADCINTFLGCLR.bit.ADCINT1 = 1; // clear ADCINT1 flag reinitialize for next SOC
        PieCtrlRegs.PIEACK.all = PIEACK_GROUP1; // acknowledge interrupt to PIE (All ADCS in group 1)
        IER |= M_INT1;
    }

    else if (obj->timeBase == DISPR_TimebaseIsPwm)
    {
        obj->ePwm->ETCLR.bit.INT = 1;  // clear INT flag for this timer
        PieCtrlRegs.PIEACK.all = PIEACK_GROUP3; // acknowledge interrupt to PIE
        IER |= M_INT3;
    }
    else
    {
        // timer
        CpuTimer0Regs.TCR.bit.TIF = 1;  // clear interrupt flag (seems to be optional)
        PieCtrlRegs.PIEACK.all = PIEACK_GROUP1; // acknowledge interrupt to PIE
        IER |= M_INT1;
    }

     // we return immediately if power-up delay has not yet expired
    if(obj->powerupCountdown > 0)
    {
        obj->powerupCountdown--;
        return;
    }

    obj->interruptNesting++;
    PLX_ASSERT(obj->interruptNesting <= 2);

    PIL_BEGIN_INT_CALL(obj->pilHandle);
    obj->thisCntdownTimestamp1 = CpuTimer0Regs.TIM.all;
    CpuTimer0Regs.TCR.bit.TIF = 1; // clear flag as we will be checking for it at next timestamp

    obj->tsk1(); // this task is always dispatched

    obj->thisCntdownTimestamp2 = CpuTimer0Regs.TIM.all;
    if(CpuTimer0Regs.TCR.bit.TIF == 1){
        obj->thisCntdownTimestamp2 = 0; // CPU timer has already bottomed out
    }

    // update worst-case values
    if(obj->cntdownTimestamp1 > obj->thisCntdownTimestamp1)
    {
        obj->cntdownTimestamp1 = obj->thisCntdownTimestamp1;
    }

    if(obj->cntdownTimestamp2 > obj->thisCntdownTimestamp2)
    {
        obj->cntdownTimestamp2 = obj->thisCntdownTimestamp2;
    }

    bool dispatchTask2 = (obj->task2DispatchTimer == 0);
    obj->task2DispatchTimer++;
    if(obj->task2DispatchTimer >= obj->task2DispatchPeriod)
    {
        obj->task2DispatchTimer = 0;
    }

    if(dispatchTask2)
    {
        PLX_ASSERT(!obj->task2Active);
        obj->task2Active = true;
        EINT; // re-enable interrupt to allow nesting
        obj->tsk2();
        DINT;
        obj->thisTask2Timestamp = obj->task2DispatchTimer;
        if(obj->task2Timestamp < obj->thisTask2Timestamp)
        {
            obj->task2Timestamp = obj->thisTask2Timestamp;
        }
        obj->task2Active = false;
    }

    DINT; // // TI expects interrupts are disabled before entering I$$REST
    obj->interruptNesting--;
}

void DISPR_background()
{
    DISPR_Obj_t *obj = (DISPR_Obj_t *)DisprHandle;

    // perform margin calculations
    DINT;
    uint32_t ts1Wc = obj->cntdownTimestamp2;
    uint16_t ts2Wc = obj->task2Timestamp;
    uint32_t ts1 = obj->thisCntdownTimestamp2;
    uint16_t ts2 = obj->thisTask2Timestamp;
    EINT;

    obj->task1Margin = (float)ts1/(float)obj->ctrlPeriod;
    obj->task2Margin = 1.0 - (float)ts2/(float)obj->task2DispatchPeriod;
    obj->task1WorstCaseMargin = (float)ts1Wc/(float)obj->ctrlPeriod;
    obj->task2WorstCaseMargin = 1.0 - (float)ts2Wc/(float)obj->task2DispatchPeriod;
}
