local Coder = { }

local T = require(Target.Name)
local U = require('CoderUtils')
local C = require('TI28xCommon')

local Registry = {
    GenericInstances = {},
    NumDigitalOutBlocks = 0,
    NumDigitalInBlocks = 0,
    GpioUsed = {},
    NumPwmBlocks = 0,
    PwmConfigs = {},
    PwmSyncConfigs = {},
    NumAdcBlocks = 0,
    AdcConfigs = {},
    AdcSocCounter = {},
    AdcSocConfigs = {},
    NumQepBlocks = 0,
    QepConfigs = {},
    NumSpiBlocks = 0,
    SpiConfigs = {},
    PwmGpio = {},
    NumReadProbes = 0,
    ReadProbeConfigs = {},
    NumOverrideProbes = 0,
    OverrideProbeConfigs = {},
    NumCalibrations = 0,
    CalibrationConfigs = {},
    PowerstageProtectionConfig = nil,
    TimerConfigs = {},
    NumTimerBlocks = 0,
    ModelTrigger = nil
}

function Coder.GetFromArrayOrScalar(field, index, majordim)
    return U.GetFromArrayOrScalar(field, index, majordim)
end

function Coder.Round(num, numDecimalPlaces)
    return U.Round(num, numDecimalPlaces)
end

function Coder.IsValidCName(name)
    return U.IsValidCName(name)
end

function Coder.RegisterGenericInstance(id)
    if Registry.GenericInstances[id] == nil then
        Registry.GenericInstances[id] = 0;
    end
    local ret = Registry.GenericInstances[id]
    Registry.GenericInstances[id] = ret+1
    return ret
end

function Coder.RegisterDigitalOutBlock()
    local ret = Registry.NumDigitalOutBlocks
    Registry.NumDigitalOutBlocks = ret+1
    return ret
end

function Coder.RegisterDigitalInBlock()
    local ret = Registry.NumDigitalInBlocks
    Registry.NumDigitalInBlocks = ret+1
    return ret
end

function Coder.RegisterPwmBlock(pwm, params, req)
    local error = C.SetPwmHardwareParams(pwm, params, req, T)
    if(error ~= nil) then
        return nil, error
    end
    local ret = Registry.NumPwmBlocks
    params["instance"] = ret
    Registry.NumPwmBlocks = ret+1
    Registry.PwmConfigs[pwm] = params
    return ret
end

function Coder.RegisterPwmSyncMux(pwm, type, unit)
    Registry.PwmSyncConfigs[pwm] = {type = '', unit = 0}
    Registry.PwmSyncConfigs[pwm]["type"] = type
    Registry.PwmSyncConfigs[pwm]["unit"] = unit
end

function Coder.RegisterAdcBlock(adc)
    if Registry.AdcConfigs[adc] == nil then
      local ret = Registry.NumAdcBlocks
      Registry.AdcConfigs[adc] = {channel = ret}
      Registry.AdcSocCounter[adc] = 0
      Registry.NumAdcBlocks = ret+1
      return ret
    else
        local adcId = 'ADC %s' % {string.char(65+adc)}
        return nil, "There can be only one %s block." % {adcId}
    end
end

function Coder.RegisterAdcSoc(adc, params, req)
    local soc = Registry.AdcSocCounter[adc]
    if soc == nil then
        return nil, "Attempting to configure SOC for uninitialized ADC."
    end
    Registry.AdcSocCounter[adc] = soc+1

    local error = C.SetAdcSocHardwareParams(adc, soc, params, req, T)
    if(error ~= nil) then
        return nil, error
    end

    local configs = Registry.AdcSocConfigs[adc]
    if(configs == nil) then
        configs = {}
        Registry.AdcSocConfigs[adc] = configs
    end
    configs[soc] = params
    return soc
end

function Coder.SetAdcSocTrigger(adc, soc, trig)
    local configs = Registry.AdcSocConfigs[adc]
    if(configs == nil) then
        return "Attempting to configure SOC trigger for uninitialized ADC."
    end
    local params = configs[soc]
    if(params == nil) then
        return "Attempting to configure SOC trigger for uninitialized SOC."
    end
    params["trig"] = trig
end

function Coder.RegisterQepBlock(qep, params, req)
    local error = C.SetQepHardwareParams(qep, params, req, T)
    if(error ~= nil) then
        return nil, error
    end
    local ret = Registry.NumQepBlocks
    params["instance"] = ret
    Registry.NumQepBlocks = ret+1
    Registry.QepConfigs[qep] = params
    return ret
end

function Coder.RegisterSpiBlock(spi, params, req)
    local error = C.SetSpiHardwareParams(spi, params, req, T)
    if(error ~= nil) then
        return nil, error
    end
    local ret = Registry.NumSpiBlocks
    params["instance"] = ret
    Registry.NumSpiBlocks = ret+1
    Registry.SpiConfigs[spi] = params
    return ret
end

function Coder.RegisterReadProbe(name, params)
    Registry.NumReadProbes = Registry.NumReadProbes+1
    Registry.ReadProbeConfigs[name] = params
end

function Coder.RegisterOverrideProbe(name, params)
    Registry.NumOverrideProbes = Registry.NumOverrideProbes+1
    Registry.OverrideProbeConfigs[name] = params
end

function Coder.RegisterCalibration(name, params)
    Registry.NumCalibrations = Registry.NumCalibrations+1
    Registry.CalibrationConfigs[name] = params
end

function Coder.RegisterPowerstageProtection(params)
    if Registry.PowerstageProtectionConfig ~= nil then
        return "There can be only one powerstage protection block."
    end
    Registry.PowerstageProtectionConfig = params
end

function Coder.RegisterTimer(params)
    if Registry.NumTimerBlocks ~= 0 then
        return "There can be only one timer block."
    end
    local error = C.SetTimerParams(0, params, T)
    if(error ~= nil) then
        return error
    end
    Registry.NumTimerBlocks = 1
    Registry.TimerConfigs[0] = params
end

function Coder.RegisterModelTrigger(trig)
    Registry.ModelTrigger = eval(trig)["modtrig"]
end

function Coder.Initialize()
    local Resources = ResourceList:new()

    T.Configure(Resources, Registry)

    return {
        HeaderDeclarations = [=[
            typedef int_fast8_t int8_t;
            typedef uint_fast8_t uint8_t;
            ]=],
        Resources = Resources,
   }
end

function Coder.Finalize()
    local Include = StringList:new()
    local Declarations = StringList:new()
    local PreInitCode = StringList:new()
    local PostInitCode = StringList:new()
    local TerminateCode = StringList:new()
    local Require = ResourceList:new()

    local error

    -- preliminary checks
    if math.floor(1/Target.Variables.SAMPLE_TIME) ~= (1/Target.Variables.SAMPLE_TIME) then
        -- this is because the dispatcher uses uint32_t variables to represent frequencies
        return "Fractional sample frequency (%f Hz) is not supported." % {1/Target.Variables.SAMPLE_TIME}
    end

    -- determine final destination for generated files
    local installDir = Target.Variables.BUILD_ROOT

    if Target.Variables.genOnly == 1 then
        installDir = Target.Variables.installDir:gsub('%"+', ''):gsub('\\+','/') -- remove quotes, make all forward slashes
        local tempDir = Target.Variables.BUILD_ROOT:gsub('%"+', ''):gsub('\\+','/') 
	if installDir..Target.Variables.BASE_NAME.."_codegen" == tempDir then -- when installDir entry is blank, returns current path
		installDir = Target.Variables.BUILD_ROOT --  use default codegen dir if installDir is blank
	end
        if not U.FileExists(installDir.."/") then
            return "The directory '%s' does not exist." % {installDir}
        end
    end

    -- PIL probes and calibrations
    local probes = io.open(Target.Variables.BUILD_ROOT .. "/probes.tmp", "w")
    io.output(probes)
    io.write("#ifndef _PROBES_CG_H_\n")
    io.write("#define _PROBES_CG_H_\n")

    io.write('#define NO_PREP_TOOL\n')
    io.write(string.format("#define CODE_GUID %s\n", U.guid()))
    io.write(string.format('#define COMPILE_TIME_DATE_STR "%s"\n', os.date("%m/%d/%Y %I:%M %p")))
    io.write('#define USER_NAME "PLECS Coder"\n');

    if (Registry.NumReadProbes > 0) or (Registry.NumOverrideProbes > 0) or (Registry.NumCalibrations > 0) then
        Declarations:append("%s_t %s;\n" % {Target.Variables.BASE_NAME, Target.Variables.BASE_NAME})

        io.write("// PIL Probes\n")
        io.write("typedef struct {\n")

        -- override probes
        for k, v in pairs(Registry.OverrideProbeConfigs) do
            local type = v['type']
            if type == nil then
                type = 'float'
            end
            io.write(string.format("  %s %s;\n", type, k))
            io.write(string.format("  %s %s_probeV;\n", type, k))
            io.write(string.format("  %s %s_probeF;\n", "int16_t", k))
        end
        -- read probes
        for k, v in pairs(Registry.ReadProbeConfigs) do
            local type = v['type']
            if type == nil then
                type = 'float'
            end
            io.write(string.format("  %s %s;\n", type, k))
        end
        -- calibrations
        for k, v in pairs(Registry.CalibrationConfigs) do
            local type = v['type']
            if type == nil then
                type = 'float'
            end
            io.write(string.format("  %s %s;\n", type, k))
        end

        io.write(string.format("} %s_t;\n",Target.Variables.BASE_NAME))
        io.write(string.format("extern %s_t %s;\n",Target.Variables.BASE_NAME, Target.Variables.BASE_NAME))
        io.write("\n");

        -- override probes
        for k, v in pairs(Registry.OverrideProbeConfigs) do
            local type = v['type']
            if type == nil then
                type = 'float'
            end
            Declarations:append("PIL_SYMBOL_DEF(%s_%s, 0, 1.0, \"\");\n" % {Target.Variables.BASE_NAME, k})
            PreInitCode:append("INIT_OPROBE(%s.%s);\n" % {Target.Variables.BASE_NAME, k})
        end
        -- read probes
        for k, v in pairs(Registry.ReadProbeConfigs) do
            local type = v['type']
            if type == nil then
                type = 'float'
            end
            Declarations:append("PIL_SYMBOL_DEF(%s_%s, 0, 1.0, \"\");\n" % {Target.Variables.BASE_NAME, k})
        end
        -- calibrations
        for k, v in pairs(Registry.CalibrationConfigs) do
            local type = v['type']
            if type == nil then
                type = 'float'
            end
            Declarations:append("PIL_SYMBOL_CAL_DEF(%s_%s, 0, 1.0, \"\", %f, %f, %f);\n" % {Target.Variables.BASE_NAME, k, v['min'], v['max'], v['val']})
            PreInitCode:append("%s.%s = %f;\n" % {Target.Variables.BASE_NAME, k, v['val']})
        end

        Include:append('pil.h')
        Include:append("%s_%s" % {Target.Variables.BASE_NAME, 'probes_cg.h'})
    end

    if(Target.Variables.EXTERNAL_MODE ~= 0) then
        io.write('#define EXTERNAL_MODE 1\n')
        io.write("#define EXTERNAL_MODE_CHECKSUM_VAR %s_checksum\n" % {Target.Variables.BASE_NAME})
    else
        io.write('#undef EXTERNAL_MODE\n')
    end
    io.write("#endif // _PROBES_CG_H_\n")
    io.close(probes)
    U.CopyTemplateFile(Target.Variables.BUILD_ROOT .. "/probes.tmp", "%s/%s_probes_cg.h" % {installDir, Target.Variables.BASE_NAME})
    os.remove(Target.Variables.BUILD_ROOT .. "/probes.tmp")

    -- external mode helper symbols
    if(Target.Variables.EXTERNAL_MODE ~= 0) then
        if(Model.NumExtModeSignals > 15) then
            return "Limit of maximal external mode signals exceeded."
        end
        Declarations:append("PIL_CONFIG_DEF(uint32_t, ExtMode_targetFloat_Size, sizeof(%s_sampleTime));\n" % {Target.Variables.BASE_NAME})
        Declarations:append("PIL_CONFIG_DEF(uint32_t, ExtMode_targetPointer_Size, sizeof(&%s_sampleTime));\n" % {Target.Variables.BASE_NAME})
        Declarations:append("PIL_CONFIG_DEF(uint32_t, ExtMode_sampleTime_Ptr, (uint32_t)&%s_sampleTime);\n" % {Target.Variables.BASE_NAME})
        Declarations:append("PIL_CONFIG_DEF(uint32_t, ExtMode_checksum_Ptr, (uint32_t)&%s_checksum);\n" % {Target.Variables.BASE_NAME})
        Declarations:append("#if defined(%s_NumTunableParameters) && (%s_NumTunableParameters >0)\n" % {Target.Variables.BASE_NAME, Target.Variables.BASE_NAME})
        Declarations:append("PIL_CONFIG_DEF(uint32_t, ExtMode_P_Ptr, (uint32_t)&%s_P);\n" % {Target.Variables.BASE_NAME})
        Declarations:append("PIL_CONFIG_DEF(uint32_t, ExtMode_P_Size, (uint32_t)%s_NumTunableParameters);\n" % {Target.Variables.BASE_NAME})
        Declarations:append("#endif\n")
        Declarations:append("#if defined(%s_NumExtModeSignals) && (%s_NumExtModeSignals > 0)\n" % {Target.Variables.BASE_NAME, Target.Variables.BASE_NAME})
        Declarations:append("PIL_CONFIG_DEF(uint32_t, ExtMode_ExtModeSignals_Ptr, (uint32_t)&%s_ExtModeSignals[0]);\n" % {Target.Variables.BASE_NAME})
        Declarations:append("PIL_CONFIG_DEF(uint32_t, ExtMode_ExtModeSignals_Size, (uint32_t)%s_NumExtModeSignals);\n" % {Target.Variables.BASE_NAME})
        Declarations:append("#endif\n")
    end
    
    error = C.ConfigureTimers(Registry, T)
    if error ~= nil then
        return error
    end

    error = C.ConfigureAdcSocTriggers(Registry, T)
    if error ~= nil then
        return error
    end

    error = C.ConfigureModelTrigger(Registry, T)
    if error ~= nil then
        return error
    end

    -- configure PWMs
    error = C.ConfigurePwms(Registry, PreInitCode, T)
    if error ~= nil then
        return error
    end

    -- configure ADCs
    error = C.ConfigureAdcs(Registry, PreInitCode, T)
    if error ~= nil then
        return error
    end

    -- configure QEPs
    error = C.ConfigureQeps(Registry, PreInitCode, T)
    if error ~= nil then
        return error
    end

    -- configure SPIs
    error = C.ConfigureSpis(Registry, PreInitCode, T)
    if error ~= nil then
        return error
    end

    -- process templates
    local dict = {}
    table.insert(dict, {before = "|>BASE_NAME<|", after = Target.Variables.BASE_NAME})
    table.insert(dict, {before = "|>TARGET_ROOT<|", after = Target.Variables.TARGET_ROOT})
    U.CopyTemplateFile(Target.Variables.TARGET_ROOT .. "/templates/settings.h", "%s/%s_settings_cg.h" % {installDir, Target.Variables.BASE_NAME}, dict)

    if Target.Variables.genOnly == 2 then
    --[[
        "One-click" build and flash
    --]]
        local codegenDir = Target.FamilySettings.ExternalTools.codegenDir:gsub('%"+', ''):gsub('\\+','/') -- remove quotes, make all forward slashes
        if not U.FileExists(codegenDir) then
            return "Codegen tools directory '%s' not found." % {codegenDir}
        end
        
        local uniflashDir = Target.FamilySettings.ExternalTools.uniflashDir:gsub('%"+', ''):gsub('\\+','/') -- remove quotes, make all forward slashes
        if not U.FileExists(uniflashDir) then
            return "Uniflash directory '%s' not found." % {uniflashDir}
        end
        
        if not U.FileExists(Target.Variables.uniflashFile) then
            return "Uniflash configuration file '%s' not found." % {Target.Variables.uniflashFile}
        end
        
        table.insert(dict, {before = "|>INSTALL_DIR<|", after = './'})
        table.insert(dict, {before = "|>BIN_DIR<|", after = "./output_%s" % {Target.Name}})
        table.insert(dict, {before = "|>CG_PATH<|", after = codegenDir})
        local srcDir = Target.Variables.TARGET_ROOT .. "/" .. Target.Name
        table.insert(dict, {before = "|>SRC_ROOT<|", after = srcDir})
                
        local uniflashExe = ""
        if Target.Variables.HOST_OS == "win" then
            uniflashExe = "%s/dslite.bat" % {uniflashDir}
        else
            uniflashExe = "%s/dslite.sh" % {uniflashDir}
        end
        if not U.FileExists(uniflashExe) then
            return "Uniflash executable '%s' not found." % {uniflashExe}
        end

        if Target.Variables.uniflashFile:sub(-#".ccxml") == ".ccxml" then
            table.insert(dict, {before = "|>CCXML_FILE<|", after = Target.Variables.uniflashFile})
        else
            table.insert(dict, {before = "|>CCXML_FILE<|", after = ""})
        end

        table.insert(dict, {before = "|>FLASH_EXE<|", after = uniflashExe})

        U.CopyTemplateFile(srcDir .. "/templates/main.mk", Target.Variables.BUILD_ROOT .. "/" .. Target.Variables.BASE_NAME .. ".mk", dict)
        U.CopyTemplateFile(srcDir .. "/templates/main.lkf", Target.Variables.BUILD_ROOT .. "/" .. Target.Variables.BASE_NAME .. ".lkf", dict)
    else
    --[[
        Generate files into CCS and build/debug from there
    --]]
        table.insert(dict, {before = "|>INSTALL_DIR<|", after = installDir})
        table.insert(dict, {before = "|>RUN_PIL_PREP<|", after = 0}) -- PIL Prep Tool will likely not be needed
        local pilPrepDir = ''
        U.CopyTemplateFile(Target.Variables.TARGET_ROOT .. "/templates/install.mk", Target.Variables.BUILD_ROOT .. "/" .. Target.Variables.BASE_NAME .. ".mk", dict)
        if Target.Variables.HOST_OS == "win" then
            table.insert(dict, {before = "|>BUILD_ROOT<|", after = Target.Variables.BUILD_ROOT:gsub('/+','\\')}) -- in batch file, we want backslashes
            table.insert(dict, {before = "|>PIL_PREP_DIR<|", after = pilPrepDir:gsub('/+','\\')})
            U.CopyTemplateFile(Target.Variables.TARGET_ROOT .. "/templates/buildsteps.bat", installDir .. "/buildsteps.bat", dict)
        else
            table.insert(dict, {before = "|>BUILD_ROOT<|", after = Target.Variables.BUILD_ROOT})
            table.insert(dict, {before = "|>PIL_PREP_DIR<|", after = pilPrepDir})
            U.CopyTemplateFile(Target.Variables.TARGET_ROOT .. "/templates/buildsteps.sh", installDir.. "/buildsteps.sh", dict)
        end

        -- helper file to specify model base-name to CCS
        U.CopyTemplateFile(Target.Variables.TARGET_ROOT .. "/templates/basename_cg.h", installDir .. "/basename_cg.h", dict)
    end

    local file = io.open(Target.Variables.BUILD_ROOT .. "/settings.tmp", "w")
    io.output(file)

    -- clock settings
    error = T.WriteClockConfiguration(io)
    if error ~= nil then
        return error
    end

    -- powerstage management
    if Registry.PowerstageProtectionConfig ~= nil then
        io.write("// powerstage management\n")
        io.write(string.format("#define DOUT_PS_ENABLE_GPIO %i \n" % {Registry.PowerstageProtectionConfig["gpio"]}))
        io.write("\n")
    end

    -- memory allocation
    io.write("// HAL object allocations\n")
    io.write("#define HAL_NUM_DOUT_CHANNELS " .. Registry.NumDigitalOutBlocks .. "\n")
    io.write("#define HAL_NUM_DIN_CHANNELS " .. Registry.NumDigitalInBlocks .. "\n")
    io.write("#define HAL_NUM_PWM_CHANNELS " .. Registry.NumPwmBlocks .. "\n")
    io.write("#define HAL_NUM_ADCS " .. Registry.NumAdcBlocks .. "\n")
    io.write("#define HAL_NUM_QEP_CHANNELS " .. Registry.NumQepBlocks .. "\n")
    io.write("#define HAL_NUM_TZ_CHANNELS 6\n") -- FIXME: No magic numbers!
    io.write("#define HAL_NUM_SPI_CHANNELS " .. Registry.NumSpiBlocks .. "\n") 
    io.write("\n")

    -- dispatcher settings
    error = C.WriteDispatcherConfiguration(io, Registry, T)
    if error ~= nil then
        return error
    end

    -- external mode comm interface
    if(Target.Variables.EXTERNAL_MODE == 1) then
        -- only SCI supported
        if(#Target.Variables.extModeSciPins ~= 2) then
            return 'Exactly two SCI pins must be specified.'
        end

        -- GPIO used for serial communication
        local sciPinset = T.GetSciPinset("%d-%d" % {Target.Variables.extModeSciPins[1], Target.Variables.extModeSciPins[2]})
        if sciPinset == nil then
            return 'SCI pin-set not supported.'
        end
        Require:add("GPIO", Target.Variables.extModeSciPins[1], "External mode communication (Rx)")
        Require:add("GPIO", Target.Variables.extModeSciPins[2], "External mode communication (Tx)")

        -- determine baud rate
        local sciBaud
        local maxRate = T.GetMaxSciBaudRate()
        local supportedRates = {256000, 128000, 115200, 57600, 38400, 19200, 14400, 9600, 4800}
        for _, rate in ipairs(supportedRates) do
            if maxRate >= rate then
                sciBaud = rate
                break
            end
        end

        if sciBaud == nil then
             return "The control task execution rate is too low to support external mode communication."
        end
        
        -- save settings
        io.write("// PIL/External mode communications\n")
        io.write("#define SCI_BAUD_RATE %d\n" % {sciBaud})
        io.write("#define SCI_PORT %d\n" % {sciPinset})
        io.write("\n")
    end

    io.close(file)
    U.CopyTemplateFile(Target.Variables.BUILD_ROOT .. "/settings.tmp", "%s/%s_settings.inc" % {installDir, Target.Variables.BASE_NAME})
    os.remove(Target.Variables.BUILD_ROOT .. "/settings.tmp")

    return {
        Declarations = Declarations,
        PreInitCode = PreInitCode,
        PostInitCode = PostInitCode,
        TerminateCode = TerminateCode,
        Include = Include,
        Require = Require
    }
end

return Coder
